/*
 * ChatClient.java
 *
 * Created on March 7, 2005, 8:05 AM
 */

package ch.epfl.lpd.ids.client;

import java.util.Observable;
import java.util.Observer;
import java.util.Hashtable;

import ch.epfl.lpd.ids.gui.ClientGUI;
import ch.epfl.lpd.ids.serialization.IMessage;
import ch.epfl.lpd.ids.utils.TopicListener;
import ch.epfl.lpd.ids.utils.ClientShutdownHook;


import javax.jms.*;
import javax.naming.*;


/**
 * This class represents a chat client. A chat client is responsible
 * for creating its GUI as well as connecting to a JMS provider.  This
 * JMS provider represents the ChatServer in some sense.  However, as
 * we are in a Publish/Subscribe system, the clients do not have a
 * list of the connected clients.
 * @author Sebastien Baehni
 */
public class ChatClient extends Observable implements IChatClient {
    
    /**
     * The {@link ch.epfl.lpd.ids.gui.ClientGUI} that represents the
     * GUI of this chat client.
     */
    private ClientGUI gui;
       
     /**
     * The username of this chat client.
     */
    protected String username = "";
    
    /**
     * Boolean used to check is this client is connected to the server
     * or not.
     */
    protected boolean connected = false;   
    
    /**
     * The connection factory that will be used to 
     * create a connection to the JMS provider.
     */
    protected ConnectionFactory cf;    
    /**
     * The connection used to create a session in order to send message.
     */
    protected Connection connection;
    /**
     * Our client is a publisher and also a subscriber.
     */   
    protected MessageProducer mp;           
    
    /**
     * The session used to send message to the JMS provider.
     */
    protected Session session;
       
    /**
     * Creates a new instance of ChatClient. This constructor
     * basically creates the {@link ch.epfl.lpd.ids.gui.ClientGUI}, and
     * initialize the MVC.
     */
    public ChatClient() {        
        this.gui = new ClientGUI(this);
        this.addObserver(this.gui);              
    }
    
    /**
     * We decouple the creation of the GUI with the fact of showing it.
     * We do this in order to call the initialize method that takes some 
     * time to connect to the JMS provider.
     */
    public void start() {
        this.gui.setVisible(true);        
    }
    
    /**
     * Used to retrieve the username of the client.
     * @return The username of the client.
     */
    public String getUsername() {
        return this.username;
    }
    
    /**
     * Method used to initialize the ChatClient. We basically
     * get the initial context used to lookup the connection factory.
     * @param contextFactory The class name of the context factory of the JMS provider.
     * @param providerURL The URL used to connect to the provider.
     * @param factoryName The topic connection factory name available on the JMS provider.
     */
    protected void initialize(String contextFactory, String providerURL, String factoryName) {
        try {
            // We set the context factory name as well as the
            // URL of the provider
            Hashtable<String,String> properties = new Hashtable<String,String>();
            properties.put(Context.INITIAL_CONTEXT_FACTORY, 
                           contextFactory);
            properties.put(Context.PROVIDER_URL, providerURL);
            
            // We now have the jndi context used to lookup the
            // topic connection factory
            Context jndiContext = new InitialContext(properties);
                       
            // We get the topic connection factory
            cf = (ConnectionFactory)jndiContext.lookup(factoryName);         
        }
        catch(NamingException ne) {
            ne.printStackTrace();  
            System.exit(0);
        }
    }
    
    /**
     * Method called by the {@link ch.epfl.lpd.ids.gui.ClientGUI} when
     * the user wants to connect to the server. This method creates a
     * {@link javax.jms.Connection}, a {@link javax.jms.Session},
     * a {@link javax.jms.MessageProducer} and a {@link javax.jms.TopicSubscriber}.
     * A chat room is created with the help of the second parameter. To 
     * create a chat room we simply create a new topic.
     * @param username The username of the client.
     * @param topicName The name of the chat room.
     * @return True if the client was able to connect, false otherwise.
     */
    public boolean connect(String username, String topicName) {
        this.username = username;
        try {
            if (!connected) {                             
                connection = cf.createConnection();
                // the session is not transacted and acknowledge automatically
                // the message it receives.
                session = connection.createSession(false, Session.AUTO_ACKNOWLEDGE);
                
                // We create the chat room. A real chat room would be a physical topic.
                // But this would imply creating topics administratively. We do not want that
                // for simplicity purpose. Due to this line, our implementation is not portable.
                // To make it portable, we should use: 
                // Topic t = (Topic)jndi.lookup("name of administrated topic");
                Topic topic = session.createTopic(topicName);
                
                // We create the publisher and the subscriber.
                mp = session.createProducer((Destination)topic);
                MessageConsumer mc = session.createConsumer((Destination)topic);
                
                // We use callbacks in order to asynchronously receive
                // messages.
                TopicListener tl = new TopicListener(this.gui);
                mc.setMessageListener(tl);
                connection.start();
                connected =true;
                return true;
            }                
        }        
        catch (JMSException jmse) {
            IMessage msg = new ch.epfl.lpd.ids.serialization.Message("A JMSException Occured While Disconnecting.", jmse.getMessage().getBytes());
            this.setChanged();            
            this.notifyObservers(msg);
            jmse.printStackTrace();
        }        
        return false;
    }
    
    /**
     * Method called by the {@link ch.epfl.lpd.ids.gui.ClientGUI} when
     * the user wants to disconnect from the JMS provider.
     * We simply close the topic connection.
     * @return True if the client was successfully disconnected, false
     * otherwise.
     */
    public boolean disconnect(String username) {
        try {           
            if (connected) {
               connection.close();             
               connected = false;
               return true;
            }
            else {
               return false;
            }
        }
        catch (JMSException jmse) {            
            IMessage msg = new ch.epfl.lpd.ids.serialization.Message("A JMSException Occured While Disconnecting.", jmse.getMessage().getBytes());
            this.setChanged();            
            this.notifyObservers(msg);
            jmse.printStackTrace();
        }
        return false;
    }
    
    /**
     * Method called by the {@link ch.epfl.lpd.ids.gui.ClientGUI} when
     * the user wants to send a message to the other chat
     * clients. This method creates a new {@link
     * ch.epfl.lpd.ids.serialization.IMessage} and sends it to the
     * JMS provider with the help of the topic session publish() method.
     * @param msg The message as a {@linnk ch.epfl.lpd.ids.serialization.IMessage}.
     */
    public void sendMessage(IMessage msg) {
        try {
            if (connected) {
                ObjectMessage om = session.createObjectMessage();
                om.setObject(msg);
                mp.send(om);
            }
        }
        catch (JMSException jse) {
            IMessage tmpMsg = new ch.epfl.lpd.ids.serialization.Message("A JMSException occured while creating/sending a message.", jse.getMessage().getBytes());
            this.setChanged();            
            this.notifyObservers(tmpMsg);
            jse.printStackTrace();
        }
    }
    
    /**
     * Method called by {@link ch.epfl.lpd.ids.gui.ClientGUI} when the
     * user close the chat GUI. We simply call the disconnect method (via
     * the client shutdown hook)
     */
    public void close() {        
        System.exit(0);
    }                       
        
    /**
     * Method that prints a help menu.
     */
    public static void printUsage() {
        System.out.println("Usage: (without the classpath)");
        System.out.println("java ch.epfl.lpd.ids.client.ChatClient <contextFactory> <providerURL> <factoryName>");
    }
    
    /**
     * Main method used to start the application.  It creates an
     * instance of a ChatClient, initialize it and start it.
     * @param args The first argument represents the context factory,
     * the second argument the URL of the JMS provider and finally the
     * last argument represents the name of the factory.
     */
    public static void main(String[] args) {
        if (args.length<3) {
            ChatClient.printUsage();
            System.exit(-1);
        }
        ChatClient cc = new ChatClient();
        cc.initialize(args[0], args[1], args[2]);
        cc.start();
        Runtime.getRuntime().addShutdownHook(new Thread(new ClientShutdownHook(cc)));   
    }
    
}
